#include<stdio.h>
#include<stdlib.h>
#include<string.h>

#include "mcd.h"
#include "util.h"
#include "dico.h"
#include "word_emb.h"


mcd *mcd_new(int nb_col)
{
  mcd *m = (mcd *)memalloc(sizeof(mcd));
  int i;
  m->nb_col = nb_col;

  for(i=0; i < FEAT_TYPE_NB; i++)
    m->type2col[i] = -1;

  m->representation = (int *)       memalloc(nb_col * sizeof(int));
  m->type =           (int *)       memalloc(nb_col * sizeof(int));
  m->type_str =       (char **)     memalloc(nb_col * sizeof(char *));
  m->filename =       (char **)     memalloc(nb_col * sizeof(char *));
  m->dico_array =     (dico **)     memalloc(nb_col * sizeof(dico *));
  m->word_emb_array = (word_emb **) memalloc(nb_col * sizeof(word_emb *));
  
  for(i=0; i < nb_col; i++){
    m->representation[i] = MCD_REPRESENTATION_NULL;
    m->type[i] = -1;
    m->type_str[i] = NULL;
    m->filename[i] = NULL;
    m->dico_array[i] = NULL;
    m->word_emb_array[i] = NULL;;
  }
  return m;
}

void mcd_free(mcd *m)
{
  int i;
  for(i=0; i < m->nb_col; i++){
    if(m->dico_array[i]) dico_free(m->dico_array[i]);
    if(m->word_emb_array[i]) word_emb_free(m->word_emb_array[i]);
    if(m->type_str[i]) free(m->type_str[i]);
  }
  free(m->representation);
  free(m->filename);
  free(m->dico_array);
  free(m->word_emb_array);
  free(m->type_str);
  free(m->type);
  free(m);
}

/* this function is used when reading an corpus file which structure is described in mcd m */
/* it returns the code associated to string str found in column col */
/* the code depends on the way the column is represented (vocabulary, embedding or integer) */ 

int mcd_get_code(mcd *m, char *str, int col){
  if(m->representation[col] == MCD_REPRESENTATION_VOCAB)
    return dico_string2int(m->dico_array[col], str);
  if(m->representation[col] == MCD_REPRESENTATION_EMB)
    return word_emb_get_code(m->word_emb_array[col], str);
  if(m->representation[col] == MCD_REPRESENTATION_INT)
    return atoi(str);
  return MCD_INVALID_VALUE;
}

/* look for the number of columns in an mcd file */

int mcd_max_column_index_in_file(char *mcd_filename)
{
  int max_col = -1;
  FILE *f = myfopen(mcd_filename, "r");
  char buffer[1000]; /* ugly */
  int column;
  char type[100];
  char representation[100];
  char filename[500]; /* ugly */
  int fields_number;
  int line_number = 0;

  while(fgets(buffer, 1000, f)){
    line_number++;
    if(feof(f)) break;
     if((buffer[0] == '\n') || (buffer[0] == '#')) continue;
    fields_number = sscanf(buffer, "%d %s %s %s", &column, type, representation, filename);
    if(fields_number != 4){
      fprintf(stderr, "line %d of mcd file %s ill formed, I'm skipping it\n", line_number, mcd_filename);
      continue;
    }
    if(column > max_col) max_col = column;
  }
  return max_col;
}


/* takes as argument an mcd structure (m) and the name of a corpus file (corpus_filename) */
/* populates the vocabularies of m with values found in corpus_filename */

void mcd_extract_dico_from_corpus(mcd *m, char *corpus_filename)
{
  int column;

  for(column=0; column < m->nb_col; column++){
    if((m->representation[column] == MCD_REPRESENTATION_VOCAB)
       /* && (strcmp(m->filename[column], "_")) */
       && (m->dico_array[column] == NULL)){
      m->dico_array[column] = dico_extract_from_corpus(corpus_filename, column, m->type_str[column]);
      fprintf(stderr, "extracting dico %s from corpus\n", m->type_str[column]);
    }
  }
}


/* takes as argument an mcd structure (m) and a dictionary vector (vocabs) */
/* links the vocabularies of m to vocabularies of vocabs (based on their names) */

void mcd_link_to_dico(mcd *m, dico_vec *vocabs, int verbose)
{
  int column;
  for(column=0; column < m->nb_col; column++){
    if((m->representation[column] == MCD_REPRESENTATION_VOCAB)
       && (!strcmp(m->filename[column], "_"))
       && (m->dico_array[column] == NULL)){
      m->dico_array[column] = dico_vec_get_dico(vocabs, m->type_str[column]);
      if(verbose) fprintf(stderr, "linking to dico %s\n", m->type_str[column]);
    }
  }
}

/* read an multi column description file and produces an mcd structure */

mcd *mcd_read(char *mcd_filename, int verbose)
{
  int column;
  char type[100];
  char representation[100];
  char filename[500]; /* ugly */
  int fields_number;
  int line_number = 0;
  char buffer[1000]; /* ugly */
  int nb_col = mcd_max_column_index_in_file(mcd_filename);
  mcd *m = mcd_new(nb_col + 1);
  FILE *f = myfopen(mcd_filename, "r");
  /* int first = 1; */

  while(fgets(buffer, 1000, f)){
    line_number++;
    if(feof(f)) break;
     if((buffer[0] == '\n') || (buffer[0] == '#')) continue;
     fields_number = sscanf(buffer, "%d %s %s %s", &column, type, representation, filename);
     if(fields_number != 4){
       /* fprintf(stderr, "line %d of mcd file %s ill formed, I'm skipping it\n", line_number, mcd_filename); */
       continue;
     }
     if(verbose) fprintf(stderr, "column = %d type = %s representation = %s filename = %s\n", column, type, representation, filename);
     m->type[column] = feat_type_string2int(type);
     m->type_str[column] = strdup(type);
     if(m->type[column] == -1){
       fprintf(stderr, "in line %d of mcd file %s invalid type, I'm skipping it\n", line_number, mcd_filename);	
       continue;
     }
    m->type2col[m->type[column]] = column;

    if(!strcmp(representation, "_")) m->representation[column] = MCD_REPRESENTATION_NULL;
    else if(!strcmp(representation, "EMB")) m->representation[column] = MCD_REPRESENTATION_EMB;
    else if(!strcmp(representation, "VOCAB")) m->representation[column] = MCD_REPRESENTATION_VOCAB;
    else if(!strcmp(representation, "INT")) m->representation[column] = MCD_REPRESENTATION_INT;
    else{ 
      fprintf(stderr, "in line %d of mcd file %s invalid mode of representation, I'm skipping it\n", line_number, mcd_filename);	  
      m->representation[column] = MCD_REPRESENTATION_NULL;
    }
    if(m->representation[column] != MCD_REPRESENTATION_NULL)
      m->filename[column] = strdup(filename);
    
    if(strcmp(m->filename[column], "_")){
      if(m->representation[column] == MCD_REPRESENTATION_EMB){
	if(verbose) fprintf(stderr, "loading word embedding %s\n", m->filename[column]);
	m->word_emb_array[column] = word_emb_load(m->filename[column]);
      }
      else if(m->representation[column] == MCD_REPRESENTATION_VOCAB){
	if(verbose) fprintf(stderr, "loading dico %s\n", m->filename[column]);
	  m->dico_array[column] = dico_read(m->filename[column], 0.5);
      }
    }
  }
  fclose(f);
  return m;
}

/* builds an mcd corresponding to the conll07 format */

mcd *mcd_build_conll07(void)
{
  mcd *m = mcd_new(8);
  m->type[0]=FEAT_TYPE_INDEX;
  m->type_str[0]=strdup("INDEX");
  m->representation[0]= MCD_REPRESENTATION_INT;
  m->filename[0] = strdup("_");
  m->type2col[FEAT_TYPE_INDEX] = 0;

  m->type[1]=FEAT_TYPE_FORM;
  m->type_str[1]=strdup("FORM");
  m->representation[1]= MCD_REPRESENTATION_VOCAB;
  m->filename[1] = strdup("_");
  m->type2col[FEAT_TYPE_FORM] = 1;
  
  m->type[2]=FEAT_TYPE_LEMMA;
  m->type_str[2]=strdup("LEMMA");
  m->representation[2]= MCD_REPRESENTATION_VOCAB;
  m->filename[2] = strdup("_");
  m->type2col[FEAT_TYPE_LEMMA] = 2;
  
  m->type[3]=FEAT_TYPE_CPOS;
  m->type_str[3]=strdup("CPOS");
  m->representation[3]= MCD_REPRESENTATION_VOCAB;
  m->filename[3] = strdup("_");
  m->type2col[FEAT_TYPE_CPOS] = 3;
  
  m->type[4]=FEAT_TYPE_POS;
  m->type_str[4]=strdup("POS");
  m->representation[4]= MCD_REPRESENTATION_VOCAB;
  m->filename[4] = strdup("_");
  m->type2col[FEAT_TYPE_POS] = 4;
  
  m->type[5]=FEAT_TYPE_FEATS;
  m->type_str[5]=strdup("FEATS");
  m->representation[5]= MCD_REPRESENTATION_VOCAB;
  m->filename[5] = strdup("_");
  m->type2col[FEAT_TYPE_FEATS] = 5;
  
  m->type[6]=FEAT_TYPE_GOV;
  m->type_str[6]=strdup("GOV");
  m->representation[6]= MCD_REPRESENTATION_INT;
  m->filename[6] = strdup("_");
  m->type2col[FEAT_TYPE_GOV] = 6;
  
  m->type[7]=FEAT_TYPE_LABEL;
  m->type_str[7]=strdup("LABEL");
  m->representation[7]= MCD_REPRESENTATION_VOCAB;
  m->filename[7] = strdup("_");
  m->type2col[FEAT_TYPE_LABEL] = 7;
  
  return m;
}

/* builds an mcd corresponding to the ifpls (index, form, pos, lemma, syntax) format */

mcd *mcd_build_ifpls(void)
{
  mcd *m = mcd_new(6);

  m->type[0]=FEAT_TYPE_INDEX;
  m->type_str[0]=strdup("INDEX");
  m->representation[0]= MCD_REPRESENTATION_INT;
  m->filename[0] = strdup("_");
  m->type2col[FEAT_TYPE_INDEX] = 0;

  m->type[1]=FEAT_TYPE_FORM;
  m->type_str[1]=strdup("FORM");
  m->representation[1]= MCD_REPRESENTATION_VOCAB;
  m->filename[1] = strdup("_");
  m->type2col[FEAT_TYPE_FORM] = 1;
  
  m->type[2]=FEAT_TYPE_POS;
  m->type_str[2]=strdup("POS");
  m->representation[2]= MCD_REPRESENTATION_VOCAB;
  m->filename[2] = strdup("_");
  m->type2col[FEAT_TYPE_POS] = 2;
  
  m->type[3]=FEAT_TYPE_LEMMA;
  m->type_str[3]=strdup("LEMMA");
  m->representation[3]= MCD_REPRESENTATION_VOCAB;
  m->filename[3] = strdup("_");
  m->type2col[FEAT_TYPE_LEMMA] = 3;
  
  m->type[4]=FEAT_TYPE_GOV;
  m->type_str[4]=strdup("GOV");
  m->representation[4]= MCD_REPRESENTATION_INT;
  m->filename[4] = strdup("_");
  m->type2col[FEAT_TYPE_GOV] = 4;
  
  m->type[5]=FEAT_TYPE_LABEL;
  m->type_str[5]=strdup("LABEL");
  m->representation[5]= MCD_REPRESENTATION_VOCAB;
  m->filename[5] = strdup("_");
  m->type2col[FEAT_TYPE_LABEL] = 5;
  
  return m;
}

mcd *mcd_read_old(char *mcd_filename, char *corpus_filename, dico_vec *vocabs)
{
  int column;
  char type[100];
  char representation[100];
  char filename[500]; /* ugly */
  int fields_number;
  int line_number = 0;
  char buffer[1000]; /* ugly */
  int nb_col = mcd_max_column_index_in_file(mcd_filename);
  mcd *m = mcd_new(nb_col + 1);
  FILE *f = myfopen(mcd_filename, "r");
  /* int first = 1; */

  while(fgets(buffer, 1000, f)){
    line_number++;
    if(feof(f)) break;
     if((buffer[0] == '\n') || (buffer[0] == '#')) continue;
     fields_number = sscanf(buffer, "%d %s %s %s", &column, type, representation, filename);
     if(fields_number != 4){
       /* fprintf(stderr, "line %d of mcd file %s ill formed, I'm skipping it\n", line_number, mcd_filename); */
       continue;
     }
     fprintf(stderr, "column = %d type = %s representation = %s filename = %s\n", column, type, representation, filename);
     m->type[column] = feat_type_string2int(type);
     if(m->type[column] == -1){
       fprintf(stderr, "in line %d of mcd file %s invalid type, I'm skipping it\n", line_number, mcd_filename);	
       continue;
     }
    m->type2col[m->type[column]] = column;
    /* m->col2type[column] = m->type[column]; */
    if(!strcmp(representation, "_")) m->representation[column] = MCD_REPRESENTATION_NULL;
    else if(!strcmp(representation, "EMB")) m->representation[column] = MCD_REPRESENTATION_EMB;
    else if(!strcmp(representation, "VOCAB")) m->representation[column] = MCD_REPRESENTATION_VOCAB;
    else if(!strcmp(representation, "INT")) m->representation[column] = MCD_REPRESENTATION_INT;
    else{ 
      fprintf(stderr, "in line %d of mcd file %s invalid mode of representation, I'm skipping it\n", line_number, mcd_filename);	  
      m->representation[column] = MCD_REPRESENTATION_NULL;
    }
    if(m->representation[column] != MCD_REPRESENTATION_NULL){
      m->filename[column] = strdup(filename);
      if(m->representation[column] == MCD_REPRESENTATION_EMB){
	fprintf(stderr, "loading word embedding %s\n", m->filename[column]);
	m->word_emb_array[column] = word_emb_load(m->filename[column]);
      }
      else if(m->representation[column] == MCD_REPRESENTATION_VOCAB){
	if(!strcmp(m->filename[column], "_")){
	  if(corpus_filename){
	    fprintf(stderr, "extracting dico %s from corpus\n", type);
	    m->dico_array[column] = dico_extract_from_corpus(corpus_filename, column, type);
	  }
	  else if(vocabs){
	    fprintf(stderr, "linking to dico %s\n", type);
	    m->dico_array[column] = dico_vec_get_dico(vocabs, type);
	  }
	  if(m->dico_array[column] == NULL)
	    fprintf(stderr, "cannot find dico %s\n", type);
	}
	else{
	  fprintf(stderr, "loading dico %s\n", m->filename[column]);
	  m->dico_array[column] = dico_read(m->filename[column], 0.5);
	}
      }
    }
  }	
  fclose(f);
  return m;
}

/* returns a dico_vec containing the different dictionnaries found in an mcd structure */

dico_vec *mcd_build_dico_vec(mcd *mcd_struct)
{
  dico_vec *dv = dico_vec_new();
  int i;
  for(i=0; i < mcd_struct->nb_col; i++){
    if(mcd_struct->dico_array[i]){
      dico_vec_add(dv, mcd_struct->dico_array[i]);
    }
  }
  return dv;
}
